#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "cmt_selector.h"
#include "cmt_tag.h"
#include "cmt_database.h"

/*----------------------------------------------------------*/
/*                                                          */
/*  Operations on Selectors                                 */
/*                                                          */
/*----------------------------------------------------------*/

/*----------------------------------------------------------*/
void Selector::unmark ()
/*----------------------------------------------------------*/
{
  if (!selected) return;

  selected = 0;
  priority = 0;
}

/*----------------------------------------------------------*/
void Selector::mark (Tag* tag)
/*----------------------------------------------------------*/
{
  if (tag == 0) return;
  if (selected) return;
  if (!tag->selected) return;

  if (name == tag->name)
    {
      selected = 1;
      priority = tag->priority;

      if (tag->tag_refs.size () > 0)
        {
          int number;

          for (number = 0; number < tag->tag_refs.size (); number++)
            {
              Tag* ref = tag->tag_refs[number];

              mark (ref);
            }
        }
    }
}

/*----------------------------------------------------------*/
void Selector::tag ()
/*----------------------------------------------------------*/
{
  int tag_number;

  if (selected) return;

  for (tag_number = 0; tag_number < Tag::tag_number (); tag_number++)
    {
      Tag* tag = Tag::tag (tag_number);

      if (tag->selected) mark (tag);
    }
}

/*----------------------------------------------------------*/
void Selector::unmark_all ()
/*----------------------------------------------------------*/
{
  static SelectorVector& Selectors = selectors ();

  int number;

  for (number = 0; number < Selectors.size (); number++)
    {
      Selector& selector = Selectors[number];

      selector.unmark ();
    }
}

/*----------------------------------------------------------*/
void Selector::mark_all (Tag* tag)
/*----------------------------------------------------------*/
{
  static SelectorVector& Selectors = selectors ();

  int number;

  Cmt::default_selector->selected = 1;
  Cmt::default_selector->priority = PriorityDefault;

  for (number = 0; number < Selectors.size (); number++)
    {
      Selector& selector = Selectors[number];

      selector.mark (tag);
    }
}

/*----------------------------------------------------------*/
void Selector::tag_all ()
/*----------------------------------------------------------*/
/*  Mark all selectors which match one of the valid tags.   */
/*----------------------------------------------------------*/
{
  static SelectorVector& Selectors = selectors ();

  int number;

  Cmt::default_selector->selected = 1;

  for (number = 0; number < Selectors.size (); number++)
    {
      Selector& selector = Selectors[number];

      selector.tag ();
    }
}

/*----------------------------------------------------------*/
Selector* Selector::create (const cmt_string& name)
/*----------------------------------------------------------*/
{
  static SelectorVector& Selectors = selectors ();

  if (Cmt::debug)
    {
      cmt_string t = "Selector::create> ";
      t += name;
      Database::dump (t, Database::key_selector);
    }

  {
    Selector* selector;
    selector = find (name);
    if (selector != 0) 
      {
        if (Cmt::debug)
          {
            cout << "Selector::create3> selector=" << selector << endl;
          }

        return (selector);
      }
  }

  Selector& selector = Selectors.add ();

  selector.name = name;
  selector.selected = 0;

  selector.tag ();

  if (Cmt::debug)
    {
      cmt_string t = "Selector::create4> ";
      t += name;
      Database::dump (t, Database::key_selector);

      cout << "Selector::create5> selector=" << &selector << endl;      
    }

  return (&selector);
}

/*----------------------------------------------------------*/
Selector* Selector::find (const cmt_string& name)
/*----------------------------------------------------------*/
{
  static SelectorVector& Selectors = selectors ();

  int number;
  Selector* result = 0;

  for (number = 0; number < Selectors.size (); number++)
    {
      Selector& selector = Selectors[number];

      if (selector.name == name)
        {
          result = &selector;
        }
    }

  return (result);
}

/*----------------------------------------------------------*/
void Selector::clear_all ()
/*----------------------------------------------------------*/
{
  static SelectorVector& Selectors = selectors ();

  if (Cmt::debug)
    {
      cmt_string t = "Selector::clear_all> ";
      Database::dump (t, Database::key_selector);
    }

  Selectors.clear ();

  if (Cmt::debug)
    {
      cmt_string t = "Selector::clear_all 2> ";
      Database::dump (t, Database::key_selector);
    }
}

/*----------------------------------------------------------*/
Selector::SelectorVector& Selector::selectors ()
/*----------------------------------------------------------*/
{
  static Database& db = Database::instance ();
  static SelectorVector& Selectors = db.selectors ();

  return (Selectors);
}

/*----------------------------------------------------------*/
Selector::Selector ()
/*----------------------------------------------------------*/
{
}

/*----------------------------------------------------------*/
Selector::~Selector ()
/*----------------------------------------------------------*/
{
}

