#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "cmt_tag.h"
#include "cmt_selector.h"
#include "cmt_database.h"

/*----------------------------------------------------------*/
/*                                                          */
/*  Operations on Tags                                      */
/*                                                          */
/*----------------------------------------------------------*/

/*----------------------------------------------------------*/
void Tag::unmark ()
/*----------------------------------------------------------*/
{
  if (!selected) return;

  selected = 0;
}

/*----------------------------------------------------------*/
void Tag::unmark_all ()
/*----------------------------------------------------------*/
{
  static TagPtrVector& Tags = tags ();

  int tag_number;

  for (tag_number = 0; tag_number < Tags.size (); tag_number++)
    {
      Tag* tag = Tags[tag_number];

      if (tag->selected) tag->unmark ();
    }
}

/*----------------------------------------------------------*/
void Tag::mark ()
/*----------------------------------------------------------*/
{
  if (selected) return;

  selected = 1;

  if (tag_refs.size () > 0)
    {
      int number;

      for (number = 0; number < tag_refs.size (); number++)
        {
          Tag* ref = tag_refs[number];

          ref->mark ();
        }
    }
}

/*----------------------------------------------------------*/
void Tag::action (const CmtSystem::cmt_string_vector& words)
/*----------------------------------------------------------*/
{
  cmt_string name;
  Tag* tag;
  Tag* ref;

  name = words[1];
  if (name == "") return;

  tag = add (name, PriorityUserTag);
  for (int i = 2; i < words.size (); i++)
    {
      name = words[i];
      if (name == "") break;
      ref = add (name, PriorityUserTag);
      tag->add_tag_ref (ref);
      if (tag->selected)
        {
          ref->mark ();
          Selector::mark_all (ref);
        }
    }
}

/*----------------------------------------------------------*/
Tag* Tag::find (const cmt_string& name)
/*----------------------------------------------------------*/
{
  static TagPtrVector& Tags = tags ();

  int tag_index;
  Tag* tag;

  if (Tags.size () == 0) return (0);

  for (tag_index = 0; tag_index < Tags.size (); tag_index++)
    {
      tag = Tags[tag_index];

      if ((tag != 0) && (tag->name == name))
        {
          return (tag);
        }
    }

  return (0);
}

/*----------------------------------------------------------*/
Tag* Tag::add (const cmt_string& name, int priority)
/*----------------------------------------------------------*/
{
  static TagPtrVector& Tags = tags ();
  static TagVector& AllTags = all_tags ();

  Tag* tag;

  if (name == "") return (0);

  if (Cmt::debug)
    {
      cmt_string t = "Tag::add> ";
      t += name;
      Database::dump (t, Database::key_tag);
    }

  tag = find (name);
  if (tag != 0)
    {
      if (priority > tag->priority) tag->priority = priority;

      if (Cmt::debug)
        {
          cout << "Tag::add3> tag=" << tag << endl;
        }

      return (tag);
    }

  Tag& tag_object = AllTags.add ();
  tag = &tag_object;
  Tags.push_back (tag);

  tag->clear ();
  tag->name = name;

  tag->selected = 0;
  tag->priority = priority;

  if (Cmt::debug)
    {
      cmt_string t = "Tag::add4> ";
      t += name;
      Database::dump (t, Database::key_tag);

      cout << "Tag::add5> tag=" << tag << endl;
    }

  return (tag);
}

/*----------------------------------------------------------*/
int Tag::tag_number ()
/*----------------------------------------------------------*/
{
  static TagPtrVector& Tags = tags ();

  return (Tags.size ());
}

/*----------------------------------------------------------*/
Tag* Tag::tag (int index)
/*----------------------------------------------------------*/
{
  static TagPtrVector& Tags = tags ();

  return (Tags[index]);
}

/*----------------------------------------------------------*/
void Tag::clear_all ()
/*----------------------------------------------------------*/
{
  static TagPtrVector& Tags = tags ();
  static TagVector& AllTags = all_tags ();

  if (Cmt::debug)
    {
      cmt_string t = "Tag::clear_all> ";
      Database::dump (t, Database::key_tag);
    }

  int tag_index;

  for (tag_index = 0; tag_index < AllTags.size (); tag_index++)
    {
      Tag& tag = AllTags[tag_index];
      tag.clear ();
    }

  Tags.clear ();
  AllTags.clear ();

  if (Cmt::debug)
    {
      cmt_string t = "Tag::clear_all 2> ";
      Database::dump (t, Database::key_tag);
    }
}

/*----------------------------------------------------------*/
Tag::TagVector& Tag::all_tags ()
/*----------------------------------------------------------*/
{
  static Database& db = Database::instance ();
  static TagVector& AllTags = db.all_tags ();

  return (AllTags);
}

/*----------------------------------------------------------*/
Tag::TagPtrVector& Tag::tags ()
/*----------------------------------------------------------*/
{
  static Database& db = Database::instance ();
  static TagPtrVector& Tags = db.tags ();

  return (Tags);
}

/*----------------------------------------------------------*/
Tag::Tag ()
/*----------------------------------------------------------*/
{
}

/*----------------------------------------------------------*/
Tag::~Tag ()
/*----------------------------------------------------------*/
{
}

/*----------------------------------------------------------*/
void Tag::clear ()
/*----------------------------------------------------------*/
{
  name = "";
  tag_refs.clear ();
  priority = PriorityUserTag;
  selected = false;
}

/*----------------------------------------------------------*/
void Tag::add_tag_ref (Tag* ref)
/*----------------------------------------------------------*/
{
  if (ref == 0) return;

  tag_refs.push_back (ref);
}

